=begin

Copyright 2011-2013, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         kmlDate.rb
Author:       Andreas Eisenbarth
Usage:        menu Plugins > 'KML Tools' > 'Add Timespan to entities'
Description:  This script allows to add begin and end time attributes to entities (primarily groups and components).
              Open the calender window from the menu and select one or more entities and choose a date.
              Dates must have the form yyy-mm-ddThh:mm (with 'T' as separator).
              The time attributes will be used by the KML-Exporter Plugin so that those entities appear
              in the Google Earth Timeline at the specified time range.
Version:      1.4.12
Date:         25.05.2013

=end
require 'sketchup.rb'



module KmlDate
@dir = File.dirname(__FILE__) # this plugin's folder



##
# Open date picker window
#
def self.run
  @model = Sketchup.active_model
  @sel = @model.selection
  # This observer brings existing time attributes into the dialog when entities are selected
  # and sets the time attributes when entities are unselected
  @selobserver = KmlDateSelObserver.new
  @sel.add_observer(@selobserver)
  # This observer is a workaround for a bug in the above observer, it does in principle the same
  @layobservers = {}
  Sketchup.active_model.layers.each{|l|
    @layobservers[l] = KmlDateLayObserver.new
    l.add_observer(@layobservers[l])
  }
  $kmlDateSelection = Set.new # set that remembers the selected entities, (even after onSelectionCleared)
  if $kmlDateDlg==nil
    $kmlDateDlg = UI::WebDialog.new("Add Timespan", false, false, 600, 200, 228, 428, true)
    html_path = File.join(@dir, "datePicker.htm")
    $kmlDateDlg.set_file(html_path)
    $kmlDateDlg.set_size(228,428)
    if Sketchup.version_number>7000000
      $kmlDateDlg.min_width = 228
      $kmlDateDlg.max_width = 228
      $kmlDateDlg.min_height = 418
      $kmlDateDlg.max_height = 428
    end
    #
    $kmlDateDlg.show {
      # javascript
      $kmlDateDlg.execute_script("document.getElementsByTagName('body')[0].style.background='"+$kmlDateDlg.get_default_dialog_color+"'")
      b = ""; e = ""
      # Use time attributes of current selection, if nothing is selected then the selection observer will do this
      if !@sel.empty?
        @sel.each{|x| $kmlDateSelection.insert x}
        b = @sel[0].get_attribute("Time","begin").to_s
        e = @sel[0].get_attribute("Time","end").to_s
      end
      $kmlDateDlg.execute_script("init()")
      $kmlDateDlg.execute_script("reset('"+b+"','"+e+"')")
    }
    $kmlDateDlg.add_action_callback("previewTimeLine") {|$kmlDateDlg, param|
      self.preview()
    }
    # When closed, save time attributes to current selection
    $kmlDateDlg.set_on_close {
      b = $kmlDateDlg.get_element_value("date-begin")
      e = $kmlDateDlg.get_element_value("date-end")
      for ent in $kmlDateSelection
        if ent.valid?
          ent.set_attribute("Time","begin",b)
          ent.set_attribute("Time","end",e)
        end
      end
      # and set maximum time range to model attributes
      # (so that it works roughly if the complete model is inserted somewhere as component)
      range = self.get_time_range()
      if (range[0]!=nil && !range[0].empty?) || (range[1]!=nil && !range[1].empty?)
        @model.set_attribute("Time","begin",range[0])
        @model.set_attribute("Time","end", (range[2]==false)? range[1]:"")
      end
      @sel.remove_observer(@selobserver)
      Sketchup.active_model.layers.each{|l|
        l.remove_observer(@layobservers[l])
      }
      $kmlDateDlg = nil
    }
  end #if $kmlDateDlg
end #def run



##
# Observe selection:
# When an entity is selected, then update the dialog with the entity's time attributes.
# When the selection is cleared, save the time values from the dialog as attributes to each selected entity.
#
class KmlDateSelObserver < Sketchup::SelectionObserver
  # when selection is changed, update selection set
  # and update dialog
  def onSelectionBulkChange(sel)
    $kmlDateSelection.clear
    sel.each{|x| $kmlDateSelection.insert x}
    if sel[0]!=nil && $kmlDateDlg!=nil && $kmlDateDlg.visible?
      b = sel[0].get_attribute("Time","begin").to_s
      e = sel[0].get_attribute("Time","end").to_s
      $kmlDateDlg.execute_script("reset('"+b+"','"+e+"')")
    end
  end
  # when selection is changed, save timespan values
  # and clear selection set
  def onSelectionCleared(sel)
    if $kmlDateDlg!=nil && $kmlDateDlg.visible?
      b = $kmlDateDlg.get_element_value("date-begin")
      e = $kmlDateDlg.get_element_value("date-end")
      for ent in $kmlDateSelection
        if ent.valid?
          ent.set_attribute("Time","begin",b)
          ent.set_attribute("Time","end",e)
        end
      end
      $kmlDateSelection.clear
      $kmlDateDlg.execute_script("reset()")
    end
  end
end



## - workaround for API bug -
# Observe when a layer is hidden that contained selected entities. Those entities are removed from the selection.
# This should have triggered 'onSelectionCleared' or 'onSelectionBulkChange' but it does not happen.
#
class KmlDateLayObserver < Sketchup::EntityObserver
  def onChangeEntity(layer)
    # if a layer has been hidden, then save the changes of time attributes
    if !layer.visible?
      #if $kmlDateDlg!=nil && $kmlDateDlg.visible?
        b = $kmlDateDlg.get_element_value("date-begin")
        e = $kmlDateDlg.get_element_value("date-end")
        for ent in $kmlDateSelection
          if ent.valid?
            ent.set_attribute("Time","begin",b)
            ent.set_attribute("Time","end",e)
          end
        end
        $kmlDateSelection.clear
        $kmlDateDlg.execute_script("reset()")
    # and - if still something is selected - reprocess the selection
        sel = Sketchup.active_model.selection
        if !sel.empty?
          sel.each{|x| $kmlDateSelection.insert x}
          if sel[0]!=nil #&& $kmlDateDlg!=nil && $kmlDateDlg.visible?
            b = sel[0].get_attribute("Time","begin").to_s
            e = sel[0].get_attribute("Time","end").to_s
            $kmlDateDlg.execute_script("reset('"+b+"','"+e+"')")
          end
        end
      #end
    end
  end
end



##
# returns number right justified and "0" added
#
def self.rjust(number,l)
  return "0"*(l-number.to_s.length)+number.to_s
end



##
# Get the maximum time range in the model (earliest begin and latest end)
#
# If no end is given, then the entity lasts until eternity. The preview() method needs
# both an begin and end in all cases. In this case it is assumed to use today's date or
# the the most future begin/end date available. This assumption is indicated by "openend".
# In one go, also remember all entities' original visibility state (only used by preview() )
#
def self.get_time_range(remember_visibility=false)
  @timeElements={} if remember_visibility==true
  earliest = nil; latest = nil; openend = false;
  today = rjust(Time.now.year,4)+"-"+rjust(Time.now.month,2)+"-"+rjust(Time.now.day,2)+"T00:00"
  @model.entities.each{|x|
    b = x.get_attribute("Time","begin")
    e = x.get_attribute("Time","end")
    if (b!=nil && !b.empty?) || (e!=nil && !e.empty?)
      @timeElements[x] = {"hidden_default"=>x.hidden?,"layer_default"=>x.layer.visible?} if !@timeElements[x] && remember_visibility==true
      earliest = b if (b!=nil && !b.empty?) && (!earliest || b<earliest)
      earliest = e if (b.nil? || b.empty?) && (e!=nil && !e.empty?) && (!earliest || e<earliest)
      latest = e if (e!=nil && !e.empty?) && (!latest || e>latest)
      # if no end date is given, entity exists until eternity
      # guess an appropriate value for latest, but signify with "openend" that it is not accurate
      openend = true if (e.nil? || e.empty?)
      latest = today if (e.nil? || e.empty?) && (b!=nil && !b.empty?) && ((latest<b && b<today) || (b<latest && latest<today))
      latest = b if (e.nil? || e.empty?) && (b!=nil && !b.empty?) && (latest<b && today<b)
    end
  }
  return [earliest, latest, openend]
end



##
# Launch/Update a timeline for previewing the time attributes.
# The timeline needs to be scaled to fit all entities' begin/end times.
#
def self.preview()
  # get the maximum time range in the model to scale the timeline
  range = self.get_time_range(true)
  if !@timeElements.empty? && range[0]!=nil && range[1]!=nil
    # if the timeline dialog is already open, update it
    if @kmlTimeLine!=nil
        @kmlTimeLine.execute_script("init('"+range[0]+"','"+range[1]+"')")
    else
    # otherwise create a new dialog
      @kmlTimeLine = UI::WebDialog.new("Timeline", false, false, 300, 300, 600, 75, true)
      html_path = File.join(@dir, "timeLine.htm")
      @kmlTimeLine.set_file(html_path)
      @kmlTimeLine.set_size(600,75)
      # dialog returns the time at the slider position, compare each entity's time with the slider's time
      @kmlTimeLine.add_action_callback("changeTime") {|@kmlTimeLine,param|
        @timeElements.each{|x,v|
          b = x.get_attribute("Time","begin")
          e = x.get_attribute("Time","end")
          if (b!=nil && !b.empty? && b>param) || (e!=nil && !e.empty? && e<param)
            x.hidden = true
            x.layer.visible = v["layer_default"]
          else
            x.hidden = false
            x.layer.visible = true
          end
        }
      }
      # restore original visibility of each entity
      @kmlTimeLine.set_on_close {
        @timeElements.each{|x,v|
          x.hidden = v["hidden_default"]
          x.layer.visible = v["layer_default"]
        }
        @kmlTimeLine = nil
      }
      # show the new dialog
      @kmlTimeLine.show {
        #@kmlTimeLine.execute_script("document.getElementsByTagName('body')[0].style.background='"+@kmlTimeLine.get_default_dialog_color+"'")
        @kmlTimeLine.execute_script("init('"+range[0]+"','"+range[1]+"')")
      }
    end
  end
end



end #module KmlDate



if( not file_loaded?(File.basename(__FILE__)) )
  cmd_add_timespan = UI::Command.new("Add Timespan to entities"){KmlDate.run}
  $kml_tools_menu = UI.menu("Plugins").add_submenu("KML Tools") if !$kml_tools_menu
  $kml_tools_menu.add_item(cmd_add_timespan)
end

file_loaded(File.basename(__FILE__))
