=begin

Copyright 2011-2013, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         kmlExport.rb
Author:       Andreas Eisenbarth
Usage:        menu Plugins > 'KML Tools' > 'Export KML File'
Description:  This plugin either exports the current selection or the complete model
              to KML Files (Keyhole Markup Language) for Google Earth.
              The following SketchUp objects can be exported:

              Groups, Components                    as  <Model> or <Multigeometry>

              grouped horizontal Faces,
              grouped extruded horizontal Faces     as  <Polygon> or <Model>

              grouped Edges, Arcs, Circles          as  <LineString> or <gx:Track> or <Model>

              ConstructionPoints, leader Texts      as  <Point>

              Images, grouped Images,
              grouped horizontal quadfaces          as  <GroundOverlay> or <Model>

              Scenes (Pages), Animation             as  <Camera>, <LookAt>, <gx:Tour>

              The model must be georeferenced first.
              Groups should not contain other groups,
              only entities on visible layers can be exported as <Model>.
Version:      1.4.12
Date:         25.05.2013

=end
# @file_dir = parent directory of kml
# @data_dir = directory for files associated with kml (contained in @file_dir)
require 'sketchup.rb'
unless Sketchup::require 'su_dynamiccomponents.rb'
  Sketchup::require 'dynamiccomponents.rb'
end



module KmlExport
@dir = File.dirname(__FILE__) # this plugin's folder



##
# Scan over the entities of the model and sort them into (possible) categories for export.
#
def self.scan_model
  @text, @cpoint, @group, @himage, @hface, @curve, _edges = [], [], [], [], [], [], []
  features = [] # contains names of categories found in the model
  @model = Sketchup.active_model
  @sel = @model.selection
  @ents = @model.entities
  @exportedMaterials = {}
  @exportedComponents = {}
  @ground = 0 # absolute elevation above sea level of origin, in [m]
  # Check Georeference
  if (Sketchup.version_number>7100000 && !@model.georeferenced? ) || @model.get_attribute("GeoReference","Longitude")==nil # SU7.1
    return UI.messagebox("The model needs to be geo-referenced.", MB_OK)
  end

  # Save selection and restore later
  # (because export as Collada only accepts selected items or complete model)
  @selection = Set.new
  @sel.each{|x| @selection.insert x}
  # Either export complete model or only selection
  elements = (@sel.empty?)? @ents : @sel
  for e in elements
   eClass = e.class
   # text
   if eClass == Sketchup::Text && e.has_leader?
     @text.push(e)
     features.push("text") unless features.include?("text")
   # construction point
   elsif eClass == Sketchup::ConstructionPoint
     @cpoint.push(e)
     features.push("cpoint") unless features.include?("cpoint")
   # curve or edges:
   # since element is only edge, collect all edges and add later to @curve
   elsif eClass == Sketchup::Edge
     _edges.push(e) # temporary array
   # horizontal image
   elsif eClass == Sketchup::Image
     if e.normal == [0,0,1]
       @himage.push({"ent" => e, "name" => e.definition.name, "description" => e.definition.description});
       features.push("himage") unless features.include?("himage");
     else
       @group.push({"ent" => e, "name" => e.definition.name, "description" => e.definition.description})
       features.push("model") unless features.include?("model")
     end
   # component
   elsif eClass == Sketchup::ComponentInstance
     @group.push({"ent" => e, "name" => (!e.name.empty?)? e.name : e.definition.name, "description" => e.definition.description})
     features.push("model") unless features.include?("model")
   # check groups for their entities
   elsif eClass == Sketchup::Group
     name = (!e.name.empty?)? e.name : e.entities.parent.name
     face = e.entities.find{|x| x.class==Sketchup::Face}
     nr_edges=e.entities.find_all{|x| x.class==Sketchup::Edge}.length
     min = e.local_bounds.min.z.round_to(10)
     max = e.local_bounds.max.z.round_to(10)
     # horizontal image
     if e.entities[0].is_a?(Sketchup::Image) && e.entities.length == 1 && e.entities[0].normal == [0,0,1]
       @himage.push({"ent" => e.entities[0], "name" => name, "description" => e.description})
       features.push("himage") unless features.include?("himage")
     # horizontal quadface
     elsif face != nil && e.entities.length == 5 && face.material != nil && face.edges.length == 4 && face.normal.transform(e.transformation.inverse) == [0,0,1]
       @himage.push({"ent" => face, "name" => name, "description" => e.description, "parent" => e})
       features.push("himage") unless features.include?("himage")
     # curve or edges
     elsif e.entities.length == nr_edges
       @curve.push({"ent" => e.entities.to_a, "name" => name, "description" => e.description, "parent" => e})
       features.push("curve") unless features.include?("curve")
     # check if horizontal face (or vertical for extrusion), count faces, count edges
     else
       ispolygon = true
       extrusion = (max-min) # maybe not necessary, only used for relativeToGround
       h_faces = []
       h_edges = []
       e.entities.each{|f|
         # ispolygon: if face is horizontal and is the highest entity of the group
         if f.is_a?(Sketchup::Face)
           n = f.normal.transform(e.transformation.inverse)
           if n.x.round_to(10) == 0 && n.y.round_to(10) == 0 && f.vertices[0].position.transform(e.transformation.inverse).z.round_to(10) == max
             h_faces.push({"ent" => f, "name" => name, "description" => e.description, "extrude" => extrusion, "parent" => e})
             h_edges += f.edges
           elsif ![0,1,-1].include? n.z.round_to(10) # reject if face is not horizontal (or vertical)
             ispolygon = false
           end
         end
       }
       h_edges.uniq!

       # polygon or extruded polygon: check if amount of entities is conform to an ideal polygon or extruded polygon
       if ispolygon == true && (e.entities.length == (h_faces.length+h_edges.length) || e.entities.length == (2*h_faces.length+4*h_edges.length) )
         @hface += h_faces
         features.push("hface") unless features.include?("hface")
       # normal group
       else
         @group.push({"ent" => e, "name" => e.name, "description" => e.description})
         features.push("model") unless features.include?("model")
       end #if ispolygon
     end #if e.entities
   end #case
  end #for elements

  # curve or edges
  if !_edges.empty?
    @curve.push({"ent" => _edges, "name" => "Curve", "description" => nil})
    features.push("curve") unless features.include?("curve");
  end
  # page
  if @model.pages.count>0
    features.push("page") unless features.include?("page");
  end
  features.push("altitudemode") unless features.empty?
  self.options_dialog(features)
end #def scan_model



##
# Ask the user which categories to export (and other options).
#
def self.options_dialog(features)
  optdlg_title = (@selection.empty?)? "Export entities to KML" : "Export selected entities to KML"
  @optdlg = UI::WebDialog.new(optdlg_title, false, false, 300, 300, 550, 400, true)
  html_path = File.join(@dir, "exportOptions.htm")
  @optdlg.set_size(550, 400)
  @optdlg.set_file(html_path)
  # callbacks
  @optdlg.add_action_callback("init"){|@optdlg, param|
    @optdlg.execute_script("document.getElementsByTagName('body')[0].style.background='"+@optdlg.get_default_dialog_color+"'")
    @optdlg.execute_script("start(['"+features.join("','")+"']);")
  }
  @optdlg.add_action_callback("windowInner") {|@optdlg, param|
    param = eval(param)
    @windowBorderWidth = (550-param[0].to_f)/2
    @windowTitlebarHeight = 400-param[1].to_f
  }
  @optdlg.add_action_callback("resizeHeight") {|@optdlg, param|
    param = eval(param)
    width = 2*@windowBorderWidth+param[0].to_f; width = 500 if width<500
    height = @windowTitlebarHeight+param[1].to_f;
    @optdlg.set_size(width, height)
  }
  @optdlg.add_action_callback("receiveOptions") { |@optdlg, params|
    options = eval(@optdlg.get_element_value("results"))
    lonlat = [@model.get_attribute("GeoReference","Longitude"), @model.get_attribute("GeoReference","Latitude")]
    self.fetch_alt(lonlat,options)
  }
  @optdlg.add_action_callback("cancel") { |@optdlg, param| @optdlg.close}
  @optdlg.show_modal
end #def options_dialog



##
# Download elevation from a web service.
#
def self.fetch_alt(lonlat,options) # latlon=[lon1,lat1]
  if Sketchup.is_online == false && @model.get_attribute("GeoReference","Altitude")==nil
      #####
      # if server does not respond
      @fetch_alt_no_error = false
      UI.start_timer(10, false) {
          if @fetch_alt_no_error==false
            puts "kmlExport.rb: no JSON returned"
            @ground =  @model.get_attribute("GeoReference","Altitude").to_f
            #self.set_geo(lonlat[0], lonlat[1], @ground)
            self.run(options)
          end
      }
      #####
      # geonames.org astergdem API (max = 20 results)
      url = "http://api.geonames.org/astergdemJSON?formatted=true&lat="
      url += lonlat[1].to_s+"&lng="+lonlat[0].to_s+"&username=aerilius&style=full&callback=returnJson"
      # callback
      @optdlg.add_action_callback("receiveJson") { |@altdlg, json|
          puts "kmlImport.rb: geonames returned #{json}"
          @fetch_alt_no_error = true
          #begin
            #raise "empty JSON" if json.empty?
            #json = eval( json.gsub(':','=>') )
            #raise "error in JSON" if json["status"]!=nil
            #@ground = json["astergdem"]-terrain_alt([0,0,0]).to_m
          if json!=nil && !json.empty? && json.to_f!=-9999
            @ground = json.to_f-terrain_alt([0,0,0]).to_m
            if @model.get_attribute("GeoReference","Altitude")==nil
              @model.set_attribute("GeoReference","Altitude", @ground)
            end
          else
          #rescue
            @ground = 0
            UI.messagebox("Loading correct elevation data failed.", MB_OK)
          end
          self.run(options)
      }
      # execute main function
      @optdlg.execute_script("getJson('"+url+"')")
  else # if is_online == true
      @ground =  @model.get_attribute("GeoReference","Altitude").to_f-terrain_alt([0,0,0]).to_m
      self.run(options)
  end
end #def fetch_alt



##
# Evaluate the options, call each category's export method and write the file.
#
def self.run(options)
# Save file dialog; create folder; create file
  suggested_name = File.basename(Sketchup.active_model.path).sub(/\.skp$/,".kml")
  @file_path = UI.savepanel("Save KML File", "#{ENV['USERPROFILE']}/", suggested_name)
  return nil if @file_path.nil?
  @file_path = File.expand_path(@file_path).unpack("U*").map{|c|c.chr}.join.sub(/(?:\.\w{0,4})?$/, ".kml")
  file_name = File.basename(@file_path, File.extname(@file_path))
  # parent directory of kml
  @file_dir = File.dirname(@file_path)
  # directory for files associated with kml
  @data_dir = @file_path.unpack("U*").map{|c|c.chr}.join.sub(/\.\w{0,4}$/, "/")
  Dir.mkdir(@data_dir) if !File.exist?(@data_dir)
  @out = File.new(@file_path, "w")
  @out.puts "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<kml xmlns=\"http:\/\/www.opengis.net\/kml\/2.2\" xmlns:gx=\"http:\/\/www.google.com\/kml\/ext\/2.2\">"
  @out.puts "<Document>\n  <name>#{file_name}<\/name>"
#  @out.puts "  <Style>\n    <ListStyle>\n      <listItemType>checkHideChildren</listItemType>\n    </ListStyle>\n  </Style>"
  # init progressbar
  n = @text.length+@cpoint.length+@curve.length+@himage.length+@hface.length+@group.length+@model.pages.count
  @optdlg.execute_script("startProgress("+n.to_s+")")
  @altitudeMode = options["kml-altitudemode"]

  # text
  if options["text"] == true
    @out.puts "<Folder>\n  <name>Texts</name>" if @text.length>5
    # export first 30 letters of text as name and the complete text as description
    eachp(@text,"exporting texts"){|x,i| self.point(x.point, x.text[/^[^\n\r]{1,30}/], x.text)}
    @out.puts "<\/Folder>" if @text.length>5
  else eachp(@text.length)
  end

  # construction point
  if options["cpoint"] == true
    @out.puts "<Folder>\n  <name>Construction Points</name>" if @cpoint.length>5
    eachp(@cpoint,"exporting points"){|x, i| self.point(x.position, "Point #{i}", nil)}
    @out.puts "<\/Folder>" if @cpoint.length>5
  else eachp(@cpoint.length)
  end

  # curve or edges
  if options["curve"] == true
    eachp(@curve,"exporting "+options["kml-curve"]+"s"){|x, i|
      x["name"] = "Curve #{i}" if x["name"].nil?
      self.plinestring(x) if options["kml-curve"] == "linestring"
      self.track(x) if options["kml-curve"] == "track"
    }
  else
    @group += @curve
  end

  # horizontal images or horizontal quadfaces
  if options["himage"] == true
    eachp(@himage,"exporting groundoverlays"){|x, i|
      x["name"] = "GroundOverlay #{i}" if x["name"].nil?
      if !(options["himage-ge"] == true && x["name"][/Google\sEarth\sSnapshot/]!=nil)
        self.groundoverlay(x)
      end
    }
  else
    @group += @himage
  end

  # polygon or extruded polygon
  if options["hface"] == true
    eachp(@hface,"exporting polygons"){|x, i|
      x["name"] = "Polygon #{i}" if x["name"].nil?
      self.ppolygon(x,options["hface-ext"])
    }
  else
    @group += @hface
  end

  # model (groups or components)
  if options["model"] == true
    dirs = {}
    eachp(@group,"exporting "+options["kml-group"]+"(s)"){|x, i|
      x["name"] = "Model #{i}" if x["name"].nil?
      if !(options["model-ge"] == true && (x["name"][/Google\sEarth\sTerrain/]!=nil || x["name"][/Google\sEarth\sSnapshot/]!=nil))
        if options["kml-group"] == "model"
          y = self.model(x) # export model and remember model filepath in dirs{}
          dirs[y] = y.sub(/.dae$/, "/")
        elsif options["kml-group"] == "multigeometry"
          self.multigeometry(x)
        end
      end
    }
    self.condense_collada(dirs) if options["model-condense"] == true && options["kml-group"] == "model"
  else eachp(@group.length)
  end

  # pages / scenes
  if options["kml-tour"] == true
    self.tour(options["kml-page"])
  # single placemarks with Camera or LookAt
  elsif options["page"] == true
    eachp(@model.pages,"exporting scenes"){|x,i|
      x["name"] = "Scene #{i}" if x.name.nil?
      @out.puts "  <Placemark>"
      @out.puts "    <name>#{x.name}</name>"
      self.time(x)
      if options["kml-page"] == "camera"
        self.camera(x)
      elsif options["kml-page"] == "lookat"
        self.lookat(x)
      end
      @out.puts "  <\/Placemark>"
      @out.puts
    }
  else eachp(@model.pages.count)
  end
  @out.puts "<\/Document>"
  @out.puts "<\/kml>"
  @out.close
  @optdlg.close if !@optdlg.nil? && @optdlg.visible?
  #Dir.foreach(@data_dir){|f|
  #  next if f=='.' or f=='..'
  #  File.delete(File.join(@data_dir, f))
  #} if Dir.entries(@data_dir)[2]
  #Dir.delete(@data_dir)
  puts "KML file successfully saved to #{@file_path}"
  # restore original selection
  @selection.each{|x| @sel.add x}
end #def run



##################################################
########## export methods for entities  ##########
##################################################


def self.model(hash)
  model = hash["ent"]
  name = hash["name"]
  description = hash["description"]

  if model.class == Sketchup::Group
    definition = model.entities.parent
  elsif (model.class == Sketchup::ComponentInstance) || (model.class == Sketchup::Image)
    definition = model.definition
  else # if something else (entities within a group)
    definition = model.parent
    model = hash["parent"] if hash["parent"]
  end
  t = model.transformation
  is_skewed = (t.xaxis.angle_between(t.yaxis).radians!=90 || t.yaxis.angle_between(t.zaxis).radians!=90 || t.xaxis.angle_between(t.zaxis).radians!=90)
# 1. Check if already exported
  href = self.check_exported(definition) # returns href of existing file or nil
  # otherwise export as a new file
  if !href || is_skewed
    href = to_filename(@data_dir, definition.name, "dae")
    # create untransformed temporary copy of group/instance and select it
    if is_skewed
      tmp_instance = @ents.add_group
      tmp_instance.entities.add_instance(definition, t)
      tmp_instance.entities[0].explode
      definition = tmp_instance.entities.parent
      t = tmp_instance.transformation
    else
      tmp_instance = @ents.add_instance(definition, Geom::Transformation.new)
    end
    # select and export selection
    @sel.clear;@sel.add tmp_instance
    options_hash = { :triangulated_faces =>  true, :doublesided_faces =>  true, :edges =>  false, :materials_by_layer =>  false, :author_attribution =>  true, :texture_maps =>  true, :selectionset_only =>  true, :preserve_instancing =>  false }
    @model.export(href, options_hash)
    @exportedComponents[definition] = href
    # delete temporary instance
    @ents.erase_entities tmp_instance
  end

# 2. Calculate location and orientation of transformed group/instance
  #t = model.transformation
  o = t.origin
  o.z = self.alt(o, @altitudeMode)
  location = @model.point_to_latlong(o).to_a
  # using additional Geom::Transformation methods from dynamic components (after comment: old code)
  heading = t.rotz # 0-X_AXIS.angle_between(t.yaxis*Z_AXIS)*180/Math::PI.to_f
  tilt = -t.rotx # -90+t.yaxis.angle_between(Z_AXIS)*180/Math::PI.to_f
  roll = t.roty # t.zaxis.angle_between(Z_AXIS*t.yaxis)*180/Math::PI.to_f-90
  # get scale
  scale = [1,1,1]
  if model.class == Sketchup::Image
    s = [model.width,model.height,0]
    us = [definition.bounds.width,definition.bounds.height,0]
  else
    s = model.scaled_size
    us = model.unscaled_size # = tmp_instance.unscaled_size
  end
  scale.each_with_index{|x,i| scale[i] = (s[i]/us[i].to_f) if (s[i]!= 0) || (us[i] != 0)}

# 3. print kml
  @out.puts "<Placemark>"
  @out.puts "  <name>#{name}<\/name>"
  @out.puts "  <description><![CDATA[#{description}]]><\/description>" unless description.nil? || description.empty?
  self.time(model)
  @out.puts "  <Model>"
  @out.puts "    <altitudeMode>#{@altitudeMode}</altitudeMode>"
  @out.puts "    <Location>"
  @out.puts "      <longitude>#{location.x}<\/longitude>"
  @out.puts "      <latitude>#{location.y}<\/latitude>"
  @out.puts "      <altitude>#{location.z}<\/altitude>"
  @out.puts "    </Location>"
  @out.puts "    <Orientation>"
  @out.puts "      <heading>#{heading}<\/heading>"
  @out.puts "      <tilt>#{tilt}<\/tilt>"
  @out.puts "      <roll>#{roll}<\/roll>"
  @out.puts "    </Orientation>"
  @out.puts "    <Scale><x>#{scale[0]}<\/x><y>#{scale[1]}<\/y><z>#{scale[2]}<\/z><\/Scale>"
  @out.puts "    <Link>"
  @out.puts "      <href>#{self.rel_path(href)}<\/href>"
  @out.puts "    <\/Link>"
  @out.puts "  <\/Model>"
  @out.puts "<\/Placemark>"
  @out.puts
  return href
end #def model



def self.multigeometry(hash) # multigeometry with polygons and linestrings
  geom = hash["ent"]
  name = hash["name"]
  description = hash["description"]
  hfaceExt = 0
  if geom.class == Sketchup::Group
    definition = geom.entities.parent
  elsif geom.class == Sketchup::ComponentInstance
    definition = geom.definition
  else # if something else (entities within a group)
    definition = geom.parent
    geom = hash["parent"] if hash["parent"]
  end
  t = geom.transformation
  allfaces = {}
  edges = []
  definition.entities.each{|x| # TODO geom.entities?
    if x.class==Sketchup::Face
      allfaces[x.material] = [] if allfaces[x.material].nil?
      allfaces[x.material].push(x)
    elsif x.class==Sketchup::Edge && x.faces.length==0 && !x.hidden?
      edges.push(x) # orphan edges
    end
  }
  allfaces.each{|m,faces|
    facecolor = self.color_to_ge(faces[0].material||geom.material)
    linecolor = self.color_to_ge(faces[0].edges[0].material) || facecolor
    @out.puts "<Placemark>"
    @out.puts "  <name>#{name}</name>"
    @out.puts "  <description>#{description}</description>"
    self.time(geom)
    if facecolor
      @out.puts "  <Style>"
      @out.puts "    <LineStyle>"
      @out.puts "      <color>#{linecolor}</color>"
      @out.puts "      <colorMode>normal</colorMode>"
      @out.puts "      <width>1</width>"
      @out.puts "    </LineStyle>"
      @out.puts "    <PolyStyle>"
      @out.puts "      <color>#{facecolor}</color>"
      @out.puts "      <colorMode>normal</colorMode>"
      @out.puts "      <fill>1</fill>"
      @out.puts "      <outline>1</outline>"
      @out.puts "    </PolyStyle>"
      @out.puts "  </Style>"
      @out.puts "  <MultiGeometry>"
    end
    faces.each{|x|
      self.polygon(x, t, hfaceExt, nil)
    }
    if !edges.empty?
      curves = self.find_connected_edges(edges)
      curves.each_with_index{|c, i|
        self.linestring(c,t)
      }
    end
    @out.puts "  </MultiGeometry>"
    @out.puts "</Placemark>"
    @out.puts
  }
end



def self.ppolygon(hash, hfaceExt) # placemark with polygon
  face = hash["ent"]
  name = hash["name"]
  parent = hash["parent"]
  t = (parent)? parent.transformation : Geom::Transformation.new
  facecolor = self.color_to_ge(face.material||parent.material)
  linecolor = self.color_to_ge(face.edges[0].material) || facecolor
  @out.puts "<Placemark>"
  @out.puts "  <name>#{name}</name>"
  self.time(parent)
  if facecolor
    @out.puts "  <Style>"
    @out.puts "    <LineStyle>"
    @out.puts "      <color>#{linecolor}</color>"
    @out.puts "      <colorMode>normal</colorMode>"
    @out.puts "      <width>1</width>"
    @out.puts "    </LineStyle>"
    @out.puts "    <PolyStyle>"
    @out.puts "      <color>#{facecolor}</color>"
    @out.puts "      <colorMode>normal</colorMode>"
    @out.puts "      <fill>1</fill>"
    @out.puts "      <outline>1</outline>"
    @out.puts "    </PolyStyle>"
    @out.puts "  </Style>"
  end
  self.polygon(face, t, hfaceExt, hash["extrude"])
  @out.puts "</Placemark>"
  @out.puts
end #def polygon



def self.polygon(face, t, hfaceExt, extrusion)
  @out.puts "  <Polygon>"
  @out.puts "    <extrude>#{(hfaceExt)? hfaceExt : 0}</extrude>"
  @out.puts "    <tessellate>0</tessellate>"
  altitudemode = (hfaceExt==true && @altitudeMode=="clampToGround")? "relativeToGround" : @altitudeMode # extrusion should override clampToGround
  @out.puts "    <altitudeMode>#{altitudemode}</altitudeMode>"
  @out.puts "    <outerBoundaryIs>"
  @out.puts "      <LinearRing>"
  @out.puts "        <coordinates>"
  outer = face.outer_loop.vertices.push(face.outer_loop.vertices.first) # to make loop closed
  outer.each{|v|
    p = v.position
    p.z = (altitudemode=="relativeToGround"&&hfaceExt==true)? extrusion : self.alt(p, altitudemode)
    p = @model.point_to_latlong(p.transform(t))
    @out.puts p.to_a.join(",")
  }
  @out.puts "        </coordinates>"
  @out.puts "      </LinearRing>"
  @out.puts "    </outerBoundaryIs>"
  @out.puts "    <innerBoundaryIs>"

  for l in face.loops
    if !l.outer?
      @out.puts "      <LinearRing>"
      @out.puts "        <coordinates>"
      inner = l.vertices.push(l.vertices.first)
      inner.each{|v|
        p = v.position
        p.z = (altitudemode=="relativeToGround"&&hfaceExt==true)? extrusion : self.alt(p, altitudemode)
        p = @model.point_to_latlong(p.transform(t))
        @out.puts p.to_a.join(",")
      }
      @out.puts "        </coordinates>"
      @out.puts "      </LinearRing>"
    end #if
  end #for
  @out.puts "    </innerBoundaryIs>"
  @out.puts "  </Polygon>"
end



def self.plinestring(hash) # placemark with linestring
  name = hash["name"]
  description = hash["description"]
  parent = hash["parent"] # may not necessarily exist
  material = (hash["ent"]!=nil)? hash["ent"][0].material : parent.material
  linecolor = self.color_to_ge(material)
  t = (parent!=nil)? parent.transformation : nil
  curves = self.find_connected_edges(hash["ent"])
  curves.each_with_index{|c, i|
    @out.puts "<Placemark>"
    @out.puts "  <name>#{name}</name>"
    self.time(parent) if parent!=nil
    if linecolor
      @out.puts "  <Style>"
      @out.puts "    <LineStyle>"
      @out.puts "      <color>#{linecolor}</color>"
      @out.puts "      <colorMode>normal</colorMode>"
      @out.puts "      <width>1</width>"
      @out.puts "    </LineStyle>"
      @out.puts "  </Style>"
    end
    self.linestring(c,t)
    @out.puts "</Placemark>"
  }
end #def linestring



def self.linestring(c,t)
  @out.puts "  <LineString>"
  @out.puts "    <altitudeMode>#{@altitudeMode}</altitudeMode>"
  @out.print "    <coordinates>"
  @out.print c.collect{|v|
    p = v.position
    p.transform!(t) if t
    p.z = self.alt(p, @altitudeMode)
    @model.point_to_latlong(p).to_a.join(",")
  }.join(" ")
  @out.puts "    </coordinates>"
  @out.puts "  </LineString>"
end #def linestring



def self.track(hash)
  name = hash["name"]
  description = hash["description"]
  ents = hash["ent"]
  parent = hash["parent"] # may not necessarily exist
  material = (!ents.nil?)? ents[0].material : parent.material
  linecolor = self.color_to_ge(material)
  t = (parent!=nil)? parent.transformation : nil
  curves = self.find_connected_edges(ents)
  curves.each_with_index{|c, i|
    @out.puts "<Placemark>"
    @out.puts "  <name>#{name} #{i}<\/name>"
    @out.puts "  <description><![CDATA[#{description}]]><\/description>" unless description.nil? || description.empty?
    self.time(parent) if parent!=nil
    if linecolor
      @out.puts "  <Style>"
      @out.puts "    <LineStyle>"
      @out.puts "      <color>#{linecolor}</color>"
      @out.puts "      <colorMode>normal</colorMode>"
      @out.puts "      <width>1</width>"
      @out.puts "    </LineStyle>"
      @out.puts "    <IconStyle>"
      @out.puts "      <scale>1</scale>"
      @out.puts "      <color>#{linecolor}</color>"
      @out.puts "    </IconStyle>"
      @out.puts "  </Style>"
    end
    @out.puts "  <gx:Track>"
    @out.puts "    <altitudeMode>#{@altitudeMode}</altitudeMode>"
    c.collect{|v|
      @out.print "<gx:coord>"
      p = v.position
      p.transform!(t) if t
      p.z = self.alt(p, @altitudeMode)
      @out.print @model.point_to_latlong(p).to_a.join(" ")
      @out.puts "    </gx:coord>"
    }
    if parent!=nil
      time_begin = parent.get_attribute("Time","begin")
      time_end = parent.get_attribute("Time","end")
      if time_begin!=nil && time_end!=nil
        @out.puts "    <when>#{time_begin}</when>"
        @out.puts "    <when>#{time_end}</when>"
      end
    end
    # interpolate angles along path
    for i in (0..c.length-2)
      p = c[i].position
      q = c[i+1].position
      dx = p.x-q.x # relative changes between two vertices
      dy = p.y-q.y
      dz = p.z-q.z
      heading = -Math.atan2(dy,dx).radians-180
      dxy = Math.sqrt(dx*dx+dy*dy)
      roll = -Math.atan2(dz,dxy).radians
      # heading, tilt, roll
      @out.puts "    <gx:angles>"+heading.to_s+" 0 "+roll.to_s+"<\/gx:angles>\n"
    end
    @out.puts "    <gx:angles>"+heading.to_s+" 0 "+roll.to_s+"<\/gx:angles>\n" # one dummy element must be added to match number of vertices
    @out.puts "  </gx:Track>"
    @out.puts "</Placemark>"
    @out.puts
  }
end #def track



def self.point(position, name, description)
  @out.puts "<Placemark>"
  @out.puts "  <name>#{name}<\/name>"
  @out.puts "  <description><![CDATA[#{description}]]><\/description>" unless description.nil? || description.empty?
  @out.puts "  <Point>"
  @out.puts "    <altitudeMode>#{@altitudeMode}</altitudeMode>"
  position.z = self.alt(position, @altitudeMode)
  p = @model.point_to_latlong(position)
  @out.puts "    <coordinates>#{p.x.to_f},#{p.y.to_f},#{p.z.to_f}<\/coordinates>"
  @out.puts "  <\/Point>"
  @out.puts "<\/Placemark>"
  @out.puts
end #def point



def self.groundoverlay(hash)
  image = hash["ent"]
  name = hash["name"]
  description = hash["description"]
  tw = Sketchup.create_texture_writer

  if image.class == Sketchup::Image
    face = image.definition.entities.find{|x| x.class == Sketchup::Face}
  else # textured quadface
    face = image # (to make the following more logical)
  end
  # create temporary group to export texture image without uv-map
  tmp_group = @model.entities.add_group
  tmp_group.material = face.material
  tw.load(tmp_group)
  image_file_name = File.basename(face.material.texture.filename)
  href = self.check_exported(face.material) # returns href of existing file or nil
  if !href # else export as a new file
    href = to_filename(@data_dir,image_file_name)
    tw.write(tmp_group, href, true)
    @exportedMaterials[face.material] = href
  end
  @ents.erase_entities(tmp_group)

  @out.puts "<GroundOverlay>"
  @out.puts "  <name>#{name}<\/name>"
  @out.puts "  <description><![CDATA[#{description}]]><\/description>" unless description.nil? || description.empty?
  self.time(image)
  @out.puts "  <Icon>"
  @out.puts "    <href>#{self.rel_path(href)}<\/href>"
  @out.puts "  <\/Icon>"
  if image.typename == "Image"
    # no skewed transformation of Sketchup::Image supported
    o = image.origin
    w = image.width
    h = image.height
    rot = image.zrotation
    c = image.bounds.center
    northwest = @model.point_to_latlong( c+[-0.5*w,0.5*h,0] )
    southeast = @model.point_to_latlong( c+[0.5*w,-0.5*h,0] )
    @out.puts "  <LatLonBox>"
    @out.puts "    <north>#{northwest.y.to_f}</north>"
    @out.puts "    <south>#{southeast.y.to_f}</south>"
    @out.puts "    <east>#{southeast.x.to_f}</east>"
    @out.puts "    <west>#{northwest.x.to_f}</west>"
    @out.puts "    <rotation>#{rot*180/Math::PI}</rotation>"
    @out.puts "  </LatLonBox>"
  else
    t = (hash["parent"])? hash["parent"].transformation : Geom::Transformation.new
    c = face.outer_loop.vertices # counter-clockwise array of vertices
    c.reverse if face.normal == [0,0,-1]
    c.collect!{|x| x = @model.point_to_latlong(x.position.transform(t)).to_a.join(",")}
    @out.puts "  <gx:LatLonQuad>"
    @out.puts "    <coordinates>#{c.join(" ")}</coordinates>"
    @out.puts "  </gx:LatLonQuad>"
  end
  @out.puts "<\/GroundOverlay>"
  @out.puts
end #def groundoverlay



def self.camera(page)
  camera = page.camera
  eye = camera.eye
  eye.z = self.alt(eye, @altitudeMode)
  eye = @model.point_to_latlong(eye)
  vector = (camera.target-camera.eye).normalize
  vectorxy = [vector.x,vector.y,0].normalize
  heading = Math.asin(vectorxy.x).radians.to_f; heading = 180-heading.to_f if vectorxy.y<0
  tilt = 90+Math.asin(vector.z).radians
  roll = (vector*camera.up).angle_between([0,0,1]).radians-90
  # fov / field of view not possible in kml
  @out.puts "<Camera>"
  @out.puts "  <longitude>#{eye.x.to_f}</longitude>"
  @out.puts "  <latitude>#{eye.y.to_f}</latitude>"
  @out.puts "  <altitude>#{eye.z.to_f}</altitude>"
  @out.puts "  <heading>#{heading}</heading>"
  @out.puts "  <tilt>#{tilt}</tilt>"
  @out.puts "  <roll>#{roll}</roll>"
  @out.puts "  <altitudeMode>#{@altitudeMode}</altitudeMode>"
  @out.puts "  <gx:ViewerOptions><option name=\"groundnavigation\" type=\"boolean\">0</option></gx:ViewerOptions>"
  @out.puts "<\/Camera>"
end #def camera



def self.lookat(page)
  camera = page.camera
  vector = (camera.target-camera.eye).normalize
  vectorxy = [vector.x,vector.y,0].normalize
  heading = Math.asin(vectorxy.x).radians.to_f; heading = 180-heading.to_f if vectorxy.y<0
  tilt = 90+Math.asin(vector.z).radians
  range = (camera.target-camera.eye).length.to_m
  t = camera.target; (vector.length = (-camera.eye.z/vector.z.to_f); t = camera.eye+vector; range = vector.length.to_f.to_m) if t.z<0 && vector.z != 0
  t.z = self.alt(t, @altitudeMode)
  target = @model.point_to_latlong(t)
  eye.z = self.alt(eye, @altitudeMode)
  # fov / field of view not possible in kml
  @out.puts "<LookAt>"
  @out.puts "  <longitude>#{target.x.to_f}</longitude>"
  @out.puts "  <latitude>#{target.y.to_f}</latitude>"
  @out.puts "  <altitude>#{target.z.to_f}</altitude>"
  @out.puts "  <heading>#{heading}</heading>"
  @out.puts "  <tilt>#{tilt}</tilt>"
  @out.puts "  <range>#{range}</range>"
  @out.puts "  <altitudeMode>#{@altitudeMode}</altitudeMode>"
  @out.puts "  <gx:ViewerOptions><option name=\"groundnavigation\" type=\"boolean\">0</option></gx:ViewerOptions>"
  @out.puts "<\/LookAt>"
end #def lookat



def self.tour(kmlPage)
  name = "Tour"
  @out.puts "<gx:Tour>"
  @out.puts "  <name>#{name}</name>"
  @out.puts "  <gx:Playlist>"
  eachp(@model.pages){|p,i|
    @out.puts "    <gx:FlyTo>"
    duration = (p.transition_time != -1)? p.transition_time : 1
    @out.puts "      <gx:duration>#{duration}</gx:duration>"
    @out.puts "      <gx:flyToMode>smooth</gx:flyToMode>"
    # export page as Camera/LookAt
    if kmlPage == "camera"
      self.camera(p)
    elsif kmlPage == "lookat"
      self.lookat(p)
    end
    @out.puts "    </gx:FlyTo>"
    wait = (p.delay_time != -1)? p.delay_time : 0
    if wait != 0
      @out.puts "    <gx:Wait>"
      @out.puts "      <gx:duration>#{wait}</gx:duration>"
      @out.puts "    </gx:Wait>"
    end
  }
  @out.puts "  </gx:Playlist>"
  @out.puts "</gx:Tour>"
  @out.puts
end #def tour



def self.time(ent)
  time_begin = ent.get_attribute("Time","begin")
  time_end = ent.get_attribute("Time","end")
  time_when = ent.get_attribute("Time","when")
  if time_begin!=nil || time_end!=nil
    @out.puts "  <TimeSpan>"
    @out.puts "    <begin>#{time_begin}</begin>" if time_begin!=nil
    @out.puts "    <end>#{time_end}</end>" if time_end!=nil
    @out.puts "  </TimeSpan>"
  elsif time_when!=nil
    @out.puts "  <TimeStamp>"
    @out.puts "    <when>#{time_when}</when>"
    @out.puts "  </TimeStamp>"
  end
end #def time



##################################################
##########       helper methods        ###########
##################################################



##
# Wrapper loop that updates progressbar
# First set the total amounts of steps to process using:
#   @optdlg.execute_script("startProgress("+n.to_s+")")
# Alternatively, the progressbar can be reset and limited to
# only the steps of this specific loop:
#   startnew=true
#   (for example if the total amount of loops is unknown)
# If Fixnum is given, then that amount of steps is skipped.
#
def self.eachp(a,processname=nil,startnew=false,&block)
  return @optdlg.execute_script("ProgressAll-="+a.to_s) if a.class==Fixnum # skip a steps
  @optdlg.execute_script("startProgress("+a.length.to_s+")") if startnew==true
  a.each_with_index{|x,i| #TODO or each{|x| if !i
    case block.arity
    when 1 then yield x if block_given?
    when 2 then yield x,i if block_given?
    end
    if i=0 && processname!=nil
      @optdlg.execute_script("progress('"+processname+"')")
    else
      @optdlg.execute_script("progress()")
    end
  }
end



##
# Converts SketchUp's rgba colors into hexadecimal aabbggrr
#
def self.color_to_ge(c) # accepts material or color
  return nil if c.nil?
  a = (c.alpha.to_f*255).to_i # get alpha value from material if possible, because color's alpha is somehow opaque
  c = c.color if c.class == Sketchup::Material
  # skp color is rgba
  skpColorArray = c.to_a
  skpColorArray[3] = a # overwrite alpha
  kmlColorArray = []
  skpColorArray.each{|x|
    # force two digits, add "0" if only one digit
    hex = ((x<16)? "0" : "") + x.to_i.to_s(16)
    # kml color is hexadecimal aabbggrr
    kmlColorArray.unshift(hex)
  }
  return kmlColorArray.to_s
end #def color_to_ge



##
# Uses an array of unsorted and eventually unconnected edges
# Returns sets of vertices of connected edges in correct order
#
def self.find_connected_edges(edges)
  counter = 0 # count remaining edges to compare
  curves = []
  v = []
  while edges.length>0
    if counter == 0
      v = edges.shift.vertices # start new sorted array v
      counter = edges.length
    end #if

    if edges.length>0 # else: single edge
      e = edges.shift # take another edge and compare
      ev = e.vertices
      o = *(ev-v) # find vertex that is not yet in in sorted array v
      if ev.include?(v.first) # if edges shares a vertex with sorted array v, add the other vertex
        o = v.last if o.nil? # only for closed rings
        v.unshift(o)
        counter = edges.length
      elsif ev.include?(v.last)
        o = v.first if o.nil? # only for closed rings
        v.push(o)
        counter = edges.length
      else # if it is not connected to the previous edge, try the next edge
        edges.push(e)
        counter -= 1
      end #if
    end
    curves.push(v) if counter == 0
  end #while
  return curves
end #def find_connected_edges



##
# Calculate altitude of a point, considering altitudeMode.
# (everything in model coordinates)
#
def self.alt(p, mode)
  alt = p.z.to_f
  ter = self.terrain_alt(p)
  if mode == "clampToGround" || mode.to_s.length==0
    return ter
  elsif mode == "relativeToGround"
    return alt-ter
  elsif mode == "absolute"
    return alt+@ground.m
  end
  # TODO: relativeToSeaFloor, clampToSeaFloor
end #def alt



##
# Get altitude from Google Earth Terrain
# (in model coordinates)
#
def self.terrain_alt(point)
  p = Geom::Point3d.new(point)
  p.z = -100000
  # raytest:
  # do NOT use [ray]=[point,vector] but point,vector,wysiwyg_flag
  # because SU<8M1 interprets two parameters also as a ray
  raytest = @model.raytest(p,[0,0,1],false)
  until raytest==nil || raytest[1][0].is_a?(Sketchup::Group) && raytest[1][0].name=="Google Earth Terrain"
      p.z = raytest[0].z
      raytest = @model.raytest(p,[0,0,1],false)
  end
  return (!raytest.nil?)? raytest[0].z : 0
end #def terrain_alt



##################################################
##########          file stuff          ##########
##################################################



##
# Convert random string into filename, remove forbidden characters
#
def self.to_filename(dir, string, ext=nil)
  # permissive: allows everything except forbidden characters of Windows&Unix file systems
  #   but ruby's File and Dir have still problems with characters like umlauts
  # string = string[/.{1,30}/].gsub(/\n|\r/,"").gsub(/(:?[\#\"\:\&\;\*\?\<\>\|\\\/]|\n|\r)+/,"_")
  # restrictive: replaces everything except standard alphabet and numbers
  if ext.nil?
    ext = string[/\.\w{1,3}$/].to_s
    string.sub!(/\.\w{1,3}$/,"")
  else
    ext = ext.sub(/^\.?/,".")
  end
  string = string[/.{1,30}/].gsub(/\n|\r/,"").gsub(/[^0-9a-zA-Z\-\_\.]+/,"_").to_s
  if !string.empty?
    base = File.join(dir, string)
  else
    base = File.join(dir, "file") # generic filename if href is empty
  end
  # detect collision of filenames and return alternative filename (numbered)
  base2 = base
  i = 0;while File.exists?(base2+ext);base2 = base+i.to_s;i += 1;end
  return base2+ext
end



##
# detect collision of filenames and return alternative filename (numbered)
#
def self.check_name(href)
  ext = href[/\.\w{1,3}$/]
  base = href.sub(/#{ext}$/, "")
  base = File.join(@data_dir, "file") if base[/[^\/]*$/].empty? # generic filename if href is empty
  base2 = base
  i = 0;while File.exists?(base2+ext);base2 = base+i.to_s;i += 1;end
  return base2+ext
end #def check_name



##
# Convert absolute file path into relative path, relative to
# - <base> directory or
# - next common directory with <file> given as second argument
#
def self.rel_path(abs,base=@file_dir)
  base=base.sub(/\/?$/,"/")
  i=0
  i+=1 until abs[i]!=base[i]
  i-=1 if i>0
  rel = abs.scan(/^.{0,#{i}}\/(.*$)/).to_s
  return (i==0)? abs : rel
end #def rel_path



##
# Check if <directory> contains a file that is identical to <file>
#
def self.file_duplicate?(file, directory)
  Dir.foreach(directory){|f|
    return f if File.size(directory+f) == File.size(file)
  }
end #def file_duplicate?



##
# Check if an identical model/texture has been exported previously,
# then just use the file path to the existing file
#
def self.check_exported(m)
  href = nil
  if m.class == Sketchup::Material
    @exportedMaterials.each{|key,value|
      href = value if m == key
    }
  elsif m.class == Sketchup::ComponentDefinition
    @exportedComponents.each{|key,value|
      href = value if m == key
    }
  end
  return href
end #def check_exported



##
# Each exported collada file has its own ressources folder.
# => Replace the many folders by a single "textures" folder that contains all ressources except duplicates
# => Make all collada files refer to the new folder
# dirs = { collada file => corresponding ressources folder }
#
def self.condense_collada(dirs)
  counter = 0 # counter for generic new filenames
  tex_dir = @data_dir+"textures/"
  Dir.mkdir(tex_dir) unless File.exist?(tex_dir)
  dirs.each{|c,d| # c = collada file; d = directory corresponding to collada file
    fileNames = Hash.new # hash to remember file name assignment
    Dir.foreach(d){|f|
      next if f == "."||f == ".."
      # check if file f is already contained in tex_dir, returns name o existing file
      newFileName = self.file_duplicate?(d+f,tex_dir)
      # if file is not yet in tex_dir, move it with new generic filename
      if newFileName == nil
        extension = f[/\.\w{0,3}$/]
        newFileName = "texture"+counter.to_s+extension
        File.move(d+f, tex_dir+newFileName)
        counter += 1
        # otherwise delete it since it is already in tex_dir
      else
        File.delete(d+f)
      end
      relSource = self.rel_path(d+f, c)
      relTarget = self.rel_path(tex_dir+newFileName, c)
      # remember filename assignment
      fileNames[relSource] = relTarget
    }
    Dir.foreach(d){|f|
      next if f=='.' or f=='..'
      File.delete(File.join(d, f))
    } if Dir.entries(d)[2]
    Dir.delete(d)
    # read Collada file and replace filenames
    collada = IO.read(c)
    fileNames.each{|old,new|
      collada.gsub!(old,new)
    }
    colladaFile = File.open(c, "w")
    colladaFile.write(collada)
    colladaFile.close
  }
  return true
end #def condense_collada



end #module KmlExport



##################################################
##########     additions to classes     ##########
##################################################



##
# Addition to Ruby File class.
# File ftools.rb, line 99
# http://www.ruby-doc.org/stdlib-1.8.6/libdoc/ftools/rdoc/classes/File.html#M000741
#
class << File
def move(from, to, verbose=false)
  #to = catname(from, to)
  $stderr.print from, " -> ", to, "\n" if verbose
  if RUBY_PLATFORM =~ /djgpp|(cyg|ms|bcc)win|mingw/ and file? to
    unlink to
  end
  fstat = stat(from)
  begin
    rename from, to
  rescue
    begin
      symlink readlink(from), to and unlink from
    rescue
      from_stat = stat(from)
      syscopy from, to and unlink from
      utime(from_stat.atime, from_stat.mtime, to)
      begin
        chown(fstat.uid, fstat.gid, to)
      rescue
      end
    end
  end
end
end #class



##
# Addition to Sketchup Image class.
# Copyright 2008, AlexM, http://forums.sketchucation.com/viewtopic.php?t=21869
#
class Sketchup::Image
   def definition
      defins = self.model.definitions
      defins.each { |d| return d if d.image? && d.instances.include?(self)}; nil
   end #def

   def transformation
      origin = self.origin
      axes = self.normal.axes
      tr = Geom::Transformation.axes(ORIGIN, axes.at(0), axes.at(1), axes.at(2))
      tr = tr*Geom::Transformation.rotation(ORIGIN, Z_AXIS, self.zrotation)
      tr = (tr*Geom::Transformation.scaling(ORIGIN, self.width/self.pixelwidth, self.height/self.pixelheight, 1)).to_a
      tr[12] = origin.x
      tr[13] = origin.y
      tr[14] = origin.z
      return Geom::Transformation.new(tr)
   end #def

   def transformation=(tr)
      self.transform! self.transformation.inverse*tr
   end
end #class



##
# Addition to Ruby Float class.
# Thanks to TIG for the code
#
class Float
  def round_to(x)
    (self*10**x).round.to_f/10**x
  end
end #class



if( not file_loaded?(File.basename(__FILE__)) )
  cmd_export_kml = UI::Command.new("Export KML File"){KmlExport.scan_model}
  $kml_tools_menu = UI.menu("Plugins").add_submenu("KML Tools") if !$kml_tools_menu
  $kml_tools_menu.add_item(cmd_export_kml)
  file_loaded(File.basename(__FILE__))
end
