=begin

Copyright 2011-2013, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         kmlImport.rb
Author:       Andreas Eisenbarth
Usage:        menu Plugins > 'KML Tools' > 'Import KML File'
Description:  It imports KML Files (Keyhole Markup Language) containing
              geo-referenced data.
              Extract kmz files (zip-archives) first. All ressources must be
              available locally, that means network links won't work.
              Supports: <Model>, <Multigeometry>, <Polygon>, <LineString>, <gx:Track>,
              <Point>, <GroundOverlay>, <PhotoOverlay>, <Camera>, <LookAt>, <gx:Tour>

              to be added: <gx:MultiTrack>,
              <gx:AnimatedUpdate>? (only AbstractViews)
Version:      1.4.12
Date:         25.05.2013

=end
# eachp" is a wrapper for "each" that updates a progressbar
# TODO: improve efficiancy of regular expressions
# bad performance for kml files > 100 kB
require 'sketchup.rb'



module KmlImport
@dir = File.dirname(__FILE__) # this plugin's folder
@debug = true


##########



def self.run(file=nil)
  @model = Sketchup.active_model
  @ents = @model.entities
  @mat = @model.materials
  @view = @model.active_view
  @pages = @model.pages
  @styles = Hash.new{{"linecolor" => nil,"facecolor" => nil,"facefill" => 1}} # holds all global styles that are identified by id
  @ground = 0 # absolute elevation above sea level of origin, in [m]

  file = UI.openpanel "Open Kml File", "c:\\", ".kml" if file.nil?
  return if file.nil?
  file = file.tr("\\","/")
  @file_dir = File.dirname(file.unpack("U*").map{|c|c.chr}.join)
  return UI.messagebox("File seems not to exist (?)", MB_OK) if !File.exists?(file)
  @kml = IO.read(file)

  # Georeference model by finding the first (random) coordinate and giving it to fetch_alt
  lonlat=[]
  if (Sketchup.version_number>7100000 && @model.georeferenced? ) || @model.get_attribute("GeoReference","Longitude")!=nil # SU7.1
    lonlat = [@model.get_attribute("GeoReference","Longitude").to_f, @model.get_attribute("GeoReference","Latitude").to_f, 0]
  else # get first coordinate
    o = @kml[/<latitude>([e0-9\.-]+)<\/latitude>(?:\s|\r|\n)*<longitude>([e0-9\.-]+)<\/longitude>(?:\s|\r|\n)*<altitude>([e0-9\.-]+)<\/altitude>/]
    o = @kml.scan(/<south>([e0-9\.-]+)<\/south>.*?<west>([e0-9\.-]+)<\/west>/m) if o.nil?
    if !o.nil? && !o.empty?
      o = o.flatten.collect!{|x| x = x.to_f}; o = o.unshift(o.slice!(1))
    else
      o = @kml.scan(/<longitude>([e0-9\.-]+)<\/longitude>(?:\s|\r|\n)*<latitude>([e0-9\.-]+)<\/latitude>(?:\s|\r|\n)*<altitude>([e0-9\.-]+)<\/altitude>/)
      o = @kml.scan(/(?:\s|>)([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?(?:\s|<)/) if o.nil? || o.empty?
      o = o.flatten.collect!{|x| x = x.to_f} if !o.nil? && !o.empty?
    end
    lonlat = o
  end
  if lonlat.nil? || lonlat.empty?
    return UI.messagebox("The file contains nothing to import.", MB_OK)
  else
    puts "Model origin: #{lonlat.join(",")}" if @debug
    return self.fetch_alt(lonlat) # lonlat=[lon1,lat1]
  end
end



##
# Download absolute elevation from a web service.
#
def self.fetch_alt(lonlat) # lonlat=[lon1,lat1]
  html_path = File.join(@dir, "fetchJson.htm")
  @dlg = UI::WebDialog.new("Progress...", false, false, 500, 400, 350, 120, true)
  @dlg.set_file(html_path)
  @dlg.set_size(350,120)
  @dlg.set_position(500,400)
  if Sketchup.version_number>7000000
    @dlg.min_width = 350
    @dlg.max_width = 350
    @dlg.min_height = 120
    @dlg.max_height = 120
  end
  @dlg.add_action_callback("init"){|@dlg, param|
      @dlg.execute_script("document.getElementsByTagName('body')[0].style.background = '"+@dlg.get_default_dialog_color+"'")
      @dlg.execute_script("document.getElementById('overlay').style.background = '"+@dlg.get_default_dialog_color+"'")
      if Sketchup.is_online == true && !lonlat.compact.empty? && @model.get_attribute("GeoReference","Altitude").nil?
          #####
          # if server does not respond
          @fetch_alt_no_error = false
          UI.start_timer(10, false) {
              if @fetch_alt_no_error==false
                  puts("kmlImport.rb: no JSON returned") if @debug
                  @ground =  @model.get_attribute("GeoReference","Altitude").to_f
                  self.set_geo(lonlat[0], lonlat[1], @ground)
                  self.run2
              end
          }
          #####
          # geonames.org astergdem API (max=20)
          url = "http://api.geonames.org/astergdemJSON?formatted=true&lat="
          url += lonlat[1].to_s+"&lng="+lonlat[0].to_s+"&username=aerilius&style=full&callback=returnJson"
puts("fetching elevation from url: #{url}") if @debug # TODO
          # execute main function
          @dlg.execute_script("getJson('"+url+"')")
      else # if is_online == false or !@model.get_attribute("GeoReference","Altitude").nil?
          @ground =  @model.get_attribute("GeoReference","Altitude").to_f-terrain_alt([0,0,0]).to_m
puts("altitude: #{@ground}") if @debug # TODO
          self.set_geo(lonlat[0], lonlat[1], @ground)
          self.run2
      end
  }
  # callback
  @dlg.add_action_callback("receiveJson") {|@dlg, json|
      puts("kmlImport.rb: geonames returned #{json}") if @debug
      @fetch_alt_no_error = true
      #begin
      if json!=nil && !json.empty? && json.to_f!=-9999
          @ground = json.to_f-terrain_alt([0,0,0]).to_m
          if (Sketchup.version_number>7100000 && !@model.georeferenced? ) || @model.get_attribute("GeoReference","Longitude")==nil # SU7.1
              # geo-reference model
              self.set_geo(lonlat[0], lonlat[1], @ground)
          elsif @model.get_attribute("GeoReference","Altitude")==nil
              # model is georefernced without altitude, add altitude only
              @model.set_attribute("GeoReference","Altitude", @ground)
          end
      else
      #rescue
          @ground = 0
          UI.messagebox("Loading correct elevation data failed.", MB_OK)
      end
      # Don't know what this timer effects, but otherwise SketchUp's dae-importer freezes
      @timer = UI.start_timer(1, false) {UI.stop_timer(@timer);self.run2}
  } # end of callback
  @dlg.show
  #return
end #def fetch_alt



##
# scan the whole file and launch import methods for each feature
#
def self.run2
  if Sketchup.version_number>7000000
    @model.start_operation("Import KML",true) # SU7.0
  else
    @model.start_operation("Import KML")
  end
  # global stylesheets
  styles = @kml.scan(/<Style\s*?id=".*?"\s*?>.*?<\/Style>/m)
  eachp(styles,"importing styles",true){|data| self.style(data.to_s) } #create Hash of styles

  stylemaps = @kml.scan(/<StyleMap\s*?id="(.*?)"\s*?>.*?<styleUrl>#(.*?)<\/styleUrl>.*?<\/StyleMap>/m)
  stylemaps.each{|data| @styles[data[0]]=@styles[data[1]]}
  placemarks = @kml.scan(/<Placemark.*?>(.*?)<\/Placemark/m) # find placemarks
  for p in placemarks # loop through all placemarks
    p = p.to_s
    name = p.scan(/<name>(.*?)<\/name/m).to_s
    description = p.scan(/<description>(.*?)<\/description/m).to_s.gsub!(/<.*?>|<!\[CDATA\[|\]\]>/,"")
    style1 = self.style(p.scan(/<Style>(.*?)<\/Style/m).to_s) # local style
    style2 = @styles[p.scan(/<StyleMap\s*?id=".*?"\s*?>.*?<styleUrl>#(.*?)<\/styleUrl>.*?<\/StyleMap>/m).to_s] # get style from StyleMap
    style2 = style2.merge(@styles[p.scan(/<styleUrl>#(.*?)<\/styleUrl/m).to_s]) # get global style
    style = style2.merge(style1)


    multi = p.scan(/(:?<MultiGeometry.*?>.*?<\/MultiGeometry>)|(:?<GeometryCollection.*?>.*?<\/GeometryCollection>)/m).to_s
    if !multi.empty?
      group = @model.entities.add_group
      pnt = multi.scan(/<Point.*?>.*?<\/Point>/m); @dlg.execute_script("startProgress("+pnt.length.to_s+")")
      eachp(pnt,"importing points",true){|data| self.point(data.to_s,(name if pnt.length<30),(description if pnt.length<30),group ) }
      lns = multi.scan(/<LineString.*?>.*?<\/LineString>/m)
      eachp(lns,"importing linestrings",true){|data| self.linestring(data.to_s,name,style,group) }
      tra = multi.scan(/<gx:Track.*?>.*?<\/gx:Track>/m)
      eachp(tra,"importing tracks",true){|data| self.track(data.to_s,style,group) }
      pol = multi.scan(/<Polygon.*?>.*?<\/Polygon>/m)
      eachp(pol,"importing polygons",true){|data| self.polygon(data.to_s,name,style,group) }
      mod = multi.scan(/<Model.*?>.*?<\/Model>/m)
      eachp(mod,"importing models",true){|data| self.model(data.to_s,name,description)}
    else
      pnt = p.scan(/<Point.*?>.*?<\/Point>/m)
      eachp(pnt,"importing points",true){|data| self.point(data.to_s,(name if pnt.length<30),(description if pnt.length<30) ) }
      lns = p.scan(/<LineString.*?>.*?<\/LineString>/m)
      eachp(lns,"importing linestrings",true){|data| self.linestring(data.to_s,name,style) }
      tra = p.scan(/<gx:Track.*?>.*?<\/gx:Track>/m)
      eachp(tra,"importing tracks",true){|data| self.track(data.to_s,style) }
      pol = p.scan(/<Polygon.*?>.*?<\/Polygon>/m)
      eachp(pol,"importing polygons",true){|data| self.polygon(data.to_s,name,style) }
      mod = p.scan(/<Model.*?>.*?<\/Model>/m)
      eachp(mod,"importing models",true){|data| self.model(data.to_s,name,description)}
    end

    cam = p.scan(/<Camera.*?>(.*?)<\/Camera>/m).to_s; camera = self.camera(cam) if !cam.empty?
    look = p.scan(/<LookAt.*?>.*?<\/LookAt>/m).to_s; camera = self.lookat(look) if !look.empty?
    if camera
      page = @pages.add(name)
      page.description = description.to_s
      page.camera.set(camera.eye, camera.target, camera.up)
    end
  end #for placemarks
  @model.materials.purge_unused

  groundoverlays = @kml.scan(/<GroundOverlay.*?>(.*?)<\/GroundOverlay/m) # find groundoverlays
  eachp(groundoverlays,"importing ground overlays",true){|data| self.groundoverlay(data.to_s) }
  photooverlays = @kml.scan(/<PhotoOverlay.*?>(.*?)<\/PhotoOverlay>/m) # find photooverlays
  eachp(photooverlays,"importing photo overlays",true){|data| self.photooverlay(data.to_s) }

  tours = @kml.scan(/<gx:Tour.*?>.*?<\/gx:Tour>/m) # find tours
  for t in tours # loop through all tours
    t = t.to_s
    name = t.scan(/<name>(.*?)<\/name>/m).to_s
    description = t.scan(/<description>(.*?)<\/description>/m).to_s.gsub!(/<.*?>|<!\[CDATA\[|\]\]>/,"").to_s
    fly = t.scan(/<gx:FlyTo.*?>.*?<\/gx:FlyTo>/m) # find flyto
    counter = 0
    for f in fly # loop through all tour
      f = f.to_s
      transition = f.scan(/<gx:duration.*?>(.*?)<\/gx:duration>/m).to_s
      cam = f.scan(/<Camera.*?<\/Camera>/m).to_s; camera = self.camera(cam) if !cam.empty?;
      look = f.scan(/<LookAt.*?<\/LookAt>/m).to_s; camera = self.lookat(look) if !look.empty?;
      #is_wait = f.scan(/<gx:Wait.*?>/m).empty?
      if camera
        page = @pages.add(name+" "+counter.to_s)
        counter += 1
        page.transition_time = transition.to_f
        page.camera.set(camera.eye, camera.target, camera.up)
      #elsif is_wait
      #  page.delay_time = transition.to_f
      end
    end #for fly
  end #for tours

  @dlg.close if !@dlg.nil? && @dlg.visible?
  if placemarks.empty? && tours.empty? && groundoverlays.empty? && photooverlays.empty?
    @model.commit_operation
    return UI.messagebox("The file contains nothing that could be imported.", MB_OK)
  else
    return @model.commit_operation
  end
end #def run



##################################################
########## import methods for entities  ##########
##################################################



def self.model(data,name,description)
  lon = data.scan(/<longitude>\s*([e0-9\.-]+)\s*<\/longitude>/).to_s.to_f
  lat = data.scan(/<latitude>\s*([e0-9\.-]+)\s*<\/latitude>/).to_s.to_f
  alt = data.scan(/<altitude>\s*([e0-9\.-]+)\s*<\/altitude>/).to_s.to_f
  mode = data.scan(/<altitudeMode>\s*(.*?)\s*<\/altitudeMode>/).to_s
  v = @model.latlong_to_point([lon,lat,alt])
  v.z = alt(v, mode)
  heading = data.scan(/<heading>\s*([e0-9\.-]+)\s*<\/heading>/).to_s.to_f
  tilt = data.scan(/<tilt>\s*([e0-9\.-]+)\s*<\/tilt>/).to_s.to_f
  roll = data.scan(/<roll>\s*([e0-9\.-]+)\s*<\/roll>/).to_s.to_f
  x = data.scan(/<x>\s*([e0-9\.-]+)\s*<\/x>/).to_s.to_f
  y = data.scan(/<y>\s*([e0-9\.-]+)\s*<\/y>/).to_s.to_f
  z = data.scan(/<z>\s*([e0-9\.-]+)\s*<\/z>/).to_s.to_f
  href = data.scan(/href>(.*?)<\/href/).to_s
  return puts("kmlImport.rb: #{href} is not a local file. Only local files can be imported.") if @debug && href[/^http|^https/]
  modhref = File.join(@file_dir, href.sub(/^\.\//,""))
  return puts("kmlImport.rb: File #{modhref} does not exist.") if @debug && (!File.exists?(modhref) || !File.extname(modhref)==".dae")
  cache = @model.entities.to_a
  success = @model.import(modhref, false)
  return if success.nil?
  imported_instance = ( @ents.to_a - cache ).first
  return if imported_instance.nil?
  h = Geom::Transformation.new([0,0,0], [0,0,1], heading.degrees)
  t = Geom::Transformation.new([0,0,0], [1,0,0], tilt.degrees)
  r = Geom::Transformation.new([0,0,0], [0,1,0], roll.degrees)
  l = Geom::Transformation.translation(v)
  s = Geom::Transformation.scaling(x,y,z)
  imported_instance.transform!(l*r*t*h*s)
  imported_instance.name = name
  imported_instance.definition.description = description.to_s
  imported_instance.definition.name = name.to_s
  return imported_instance
end #def model



def self.polygon(data,name,style,context=@ents.add_group)
  altOffSet = data.scan(/<gx:altitudeOffset>(?:\s|\r|\n)*([e0-9\.-]+)(?:\s|\r|\n)*<\/gx:altitudeOffset>/m).to_s.to_f
  extrude = data.scan(/<extrude>(?:\s|\r|\n)*([01])*<\/extrude>/m).to_s.to_i
  extrusion = 0
  mode = data.scan(/<altitudeMode>\s*(.*?)\s*<\/altitudeMode>/).to_s
  outer = data.scan(/<outerBoundaryIs>(?:\s|\r|\n)*<LinearRing>(?:\s|\r|\n)*<coordinates>(.*?)<\/coordinates>(?:\s|\r|\n)*<\/LinearRing>(?:\s|\r|\n)*<\/outerBoundaryIs>/m).to_s

  # outerLoop:
  outerLoop = outer.scan(/([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/m).compact
  max = Float::MIN # highest altitude in polygon, necessary for extruded polygons in order to touch the ground
  min_ter = 0.0 # lowest terrain altitude, necessary for extruded polygons in order to touch the ground
  mid = 0.0 # the average terrain altitude + extrusion; sum up all z values and later divide by amount of summands; TODO: average or altitude at center?
  outerLoop = outerLoop.collect{|x|
    x.collect!{|y| y.to_f }
    x = @model.latlong_to_point(x) # convert into model coordinates
    x.z = alt(x, mode) + altOffSet.m # and set altitude according to altitudeMode
    max = x.z if x.z > max
    tmp_ter = self.terrain_alt(x)
    mid += x.z
    min_ter = tmp_ter if tmp_ter < min_ter
    x = x.to_a
  }.uniq.compact
  return if outerLoop.length < 3
  mid = mid/(1+outerLoop.length).to_f # sum of altitudes divided by amount of summands, gives average altitude

  plane = self.points_to_plane(outerLoop)
  outerLoop.collect!{|x|
    if extrude == 1 && style["facefill"]!=0
      x.z = mid # extruded polygon should be horizontal and not follow terrain
      x
    else
      x = x.project_to_plane(plane) # make x fit into plane to better handle unprecisions
    end
  }

  # innerLoop: analogue, all in one step
  inner = data.scan(/<innerBoundaryIs>(.*?)<\/innerBoundaryIs>/m).to_s
  if !inner.empty?
    innerLoops = inner.scan(/<LinearRing>(?:\s|\r|\n)*<coordinates>.*?<\/coordinates>(?:\s|\r|\n)*<\/LinearRing>/m)
    innerLoops.collect!{|x|
      x = x.scan(/([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/m).compact
      x.collect!{|y|
        y.collect!{|z| z.to_f }
        y = @model.latlong_to_point(y).to_a
        y.z = alt(y, mode) + altOffSet.m
        y.z = mid if extrude == 1 && style["facefill"]!=0
        y = y.project_to_plane(plane)
      }
      x.uniq.compact
    }
  else
    innerLoops = []
  end

  # draw outer loop
  cache = context.entities.to_a # array that contains all created entities
  context.name = name if context.name.empty?
  if style["facefill"]==0
    outerLoop.push(outerLoop.first)
    context.entities.add_edges(outerLoop)
  else
    outerLoop.uniq! # prevent error of duplicate points in add_face
    # eventual problem: duplicate points cannot just be removed without affecting the polygon
    # FIXME if needed:    outerLoop = self.split_uniq(outerLoop); outerLoop.each{ # splits array in smaller pieces, each without duplicates
    if outerLoop.length>=3
      begin
        face = context.entities.add_face(outerLoop)
        face.reverse! if face.normal.z==-1
      rescue; puts("KmlImport Error: one polygon was not planar") if @debug
      end
    end
  end

  # draw innerloops
  if style["facefill"]==0
    innerLoops.each {|l|
      l.push(l.first)
      context.entities.add_edges(l)
    }
  else
    innerLoops.each {|l|
      # FIXME if needed:    l = self.split_uniq(l); l.each{ # prevent error of duplicate points in add_face
      l.uniq!
      # next if !l.empty? # FIXME
      if l.length>=3
        begin
          f = context.entities.add_face(l)
          context.entities.erase_entities(f)
        rescue; puts("KmlImport Error: one polygon was not planar") if @debug
        end
      end
    }
  end
  # extrude from polygon to lowest terrain altitude
  extrusion = (mid-min_ter).abs
  if extrude == 1 && style["facefill"]!=0
    face.pushpull(-extrusion)
  end
  # colorize
  created_entities = context.entities.to_a-cache
  created_entities.each{|x|
    if x.class==Sketchup::Edge
      x.material = style["linecolor"]
    elsif x.class==Sketchup::Face
      x.material = style["facecolor"]
      x.back_material = style["facecolor"]
      # TODO: is this redundant?
      # context.entities.erase_entities(x) if style["facefill"] == 0 # delete face if face shouldn't be filled
    end
  }
  @model.rendering_options["EdgeColorMode"] = 0 unless style["linecolor"].nil? # turn colored edges style on
end #def polygon



def self.linestring(data,name,style,context=@ents.add_group)
  altOffSet = data.scan(/<gx:altitudeOffset>(?:\s|\r|\n)*([e0-9\.-]+)(?:\s|\r|\n)*<\/gx:altitudeOffset>/m).to_s.to_f
  mode = data.scan(/<altitudeMode>\s*(.*?)\s*<\/altitudeMode>/).to_s
  mode = "clampToGround" if data.scan(/<tesselated>\s*(.*?)\s*<\/tesselated>/).to_s.to_i == 1
  c = data.scan(/([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/m).compact # find list of coordinates
  c = c.collect{|x|
    x.collect!{|y| y.to_f } # turn strings into floating numbers
    x = @model.latlong_to_point(x)
    x.z = alt(x, mode) + altOffSet.m;
    x = x.to_a
  }
  return puts("KmlImport Error: Linestring has not enough non-identical points (#{c.inspect})") if @debug && c.length < 2
  # create group with curve from points
  context.entities.add_curve c
  context.name = name #if context.name.empty?
  context.material = style["linecolor"]
  @model.rendering_options["EdgeColorMode"] = 0 unless style["linecolor"].nil? # turn colored edges style on
end #def linestring



def self.track(data,style,context=@ents.add_group)
  name = data.scan(/name>(.*?)<\/name/m).to_s
  mode = data.scan(/<altitudeMode>\s*(.*?)\s*<\/altitudeMode>/).to_s
  mod = data.scan(/<Model.*?>.*?<\/Model>/m).to_s
  c = data.scan(/<gx:coord>([e0-9\.-]+)\s([e0-9\.-]+)\s?([e0-9\.-]+)?<\/gx:coord>/m) # find list of coordinates #!!!!! [0]
  c = c.collect{|x|
    x.collect!{|y| y.to_f } # turn strings into floating numbers
    x = @model.latlong_to_point(x)
    x.z = alt(x, mode);
    x = x.to_a
  }
  # create group with curve from points
  context.entities.add_curve c
  context.name = name if context.name.empty?
  context.material = style["linecolor"]
  @model.rendering_options["EdgeColorMode"] = 0 unless style["linecolor"].nil? # turn colored edges style on
  # import model if available
  if mod.length != 0
    imported = self.model(mod,name)
    l = Geom::Transformation.translation(pt[0])
    imported.transform!(l)
  end
end #def track



def self.point(data,name,description,context=@model.entities)
  mode = data.scan(/<altitudeMode>\s*(.*?)\s*<\/altitudeMode>/).to_s
  c = data.scan(/([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/m)[0]
  c.collect!{|x| x.to_f } # turn strings into floating numbers
  c = @model.latlong_to_point(c)
  c.z = alt(c, mode)

  description.gsub!(/<.*?>|<!\[CDATA\[|\]\]>/,"") if !description.nil?
  # limit for not adding to long description texts
  # better: if Sketchup::Text had a Text.height= method
  description = description.slice(/^[^\n\r]{1,50}/) if !description.nil?
  if !description.nil? && !description.empty?
    context.add_text description, c, [1,1,1]
    context.add_cpoint c
  elsif !name.empty?
    context.add_text name, c, [1,1,1]
    context.add_cpoint c
  else
    context.add_cpoint c
  end
end #def point



def self.groundoverlay(g)
  name = g.scan(/name>(.*?)<\/name/m).to_s
  description = g.scan(/description>(.*?)<\/description/m).to_s.gsub!(/<.*?>|<!\[CDATA\[|\]\]>/,"").to_s
  href = g.scan(/href>(.*?)<\/href/m).to_s
  return puts("kmlImport.rb: #{href} is not a local file. Only local files can be imported.") if @debug && href[/^http|^https/]
  imghref = File.join(@file_dir, href.sub(/^\.\//,""))
  return puts("kmlImport.rb: File #{imghref} does not exist.") if @debug && !File.exists?(imghref)
  # case 1: latlongquad with 4 corner points
  if !g.scan(/gx:LatLonQuad/).empty?
    data = g.scan(/<gx:LatLonQuad.*?coordinates>\s*(.*?)\s*<\/coordinates.*?<\/gx:LatLonQuad/m).to_s
    c = data.scan(/([e0-9\.-]+),([e0-9\.-]+)/)
    c.collect!{|x| x = @model.latlong_to_point( [x[0].to_f, x[1].to_f, 0] ); x[2] = 0; x }
    c.unshift(c.pop) # last element to begin
    group = @ents.add_group
    group.name = name
    group.description = description
    face = group.entities.add_face(*c)
    face.reverse! if face.normal.z < 0
    m = @mat.add name
    m.texture = imghref
    face.material = m
    pt_array = [c[0],[0,1,0],c[1],[0,0,0],c[2],[1,0,0],c[3],[1,1,0]]
    face.position_material m, pt_array, true
  # case 2: N,S,E,W,rotation
  else
    north = g.scan(/north>(.*?)<\/north/m).to_s.to_f
    south = g.scan(/south>(.*?)<\/south/m).to_s.to_f
    east = g.scan(/east>(.*?)<\/east/m).to_s.to_f
    west = g.scan(/west>(.*?)<\/west/m).to_s.to_f
    rot = g.scan(/rotation>(.*?)<\/rotation/m).to_s.to_f
    # calculate width/height from 3 points
    p1 = @model.latlong_to_point([west,south]) #@model.latlong_to_point([south,west]) #
    p2 = @model.latlong_to_point([west,north]) #@model.latlong_to_point([north,west]) #
    p3 = @model.latlong_to_point([east,north]) #@model.latlong_to_point([north,east]) #
    center = @model.latlong_to_point([(west+east)/2,(north+south)/2,0]) #@model.latlong_to_point([(north+south)/2,(west+east)/2,0]) #
    w = p2.distance p3
    h = p2.distance p1

    image = @ents.add_image imghref, p1, w, h
    rotCorr = -Math.acos((p2.y-p1.y)/h)
    rotationCorr = Geom::Transformation.rotation(p1, [0,0,1], rotCorr)
    image.transform! rotationCorr if image!=nil
    rotation = Geom::Transformation.rotation(center, [0,0,1], rot.degrees)
    image.transform! rotation if image!=nil
    image.definition.description = description.to_s if image!=nil
  end
end #def groundoverlay



def self.photooverlay(p)
  name = p.scan(/name>(.*?)<\/name/m).to_s
  href = p.scan(/(?:<Style>.*?<\/Style>.*?)?<Icon>.*?<href>(.*?)<\/href>.*?<\/Icon>/m).to_s
  return puts("kmlImport.rb: #{href} is not a local file. Only local files can be imported.") if @debug && href[/^http|^https/]
  imghref = File.join(@file_dir, href.sub(/^\.\//,""))
  puts("kmlImport.rb: File #{imghref} does not exist.") if @debug && !File.exists?(imghref)
  leftFov = p.scan(/leftFov>(.*?)<\/leftFov/m).to_s.to_f
  rightFov = p.scan(/rightFov>(.*?)<\/rightFov/m).to_s.to_f
  bottomFov = p.scan(/bottomFov>(.*?)<\/bottomFov/m).to_s.to_f
  topFov = p.scan(/topFov>(.*?)<\/topFov/m).to_s.to_f
  cam = p.scan(/<Camera>.*?<\/Camera>/m).to_s; camera = self.camera(cam) if cam.length != 0;
  look = p.scan(/<LookAt>.*?<\/LookAt>/m).to_s; camera = self.lookat(look) if look.length != 0;
  if camera.nil?; camera = Sketchup::Camera.new; end
  camera.fov = (-leftFov+rightFov-bottomFov+topFov)/2 # top+bottom  <=>  left+right ?
  # camera.aspect_ratio = (leftFov+rightFov)/(bottomFov+topFov) ?
  page = @pages.add_matchphoto_page(imghref, camera, name)
end #def photooverlay



def self.camera(cam)
  mode = cam.scan(/<altitudeMode>\s*(.*?)\s*<\/altitudeMode>/).to_s
  lon = cam.scan(/<longitude>\s*([e0-9\.-]+)\s*<\/longitude>/m).to_s.to_f
  lat = cam.scan(/<latitude>\s*([e0-9\.-]+)\s*<\/latitude>/m).to_s.to_f
  alt = cam.scan(/<altitude>\s*([e0-9\.-]+)\s*<\/altitude>/m).to_s.to_f
  heading = cam.scan(/<heading>\s*([e0-9\.-]+)\s*<\/heading>/m).to_s.to_f
  tilt = cam.scan(/<tilt>\s*([e0-9\.-]+)\s*<\/tilt>/m).to_s.to_f
  roll = cam.scan(/<roll>\s*([e0-9\.-]+)\s*<\/roll>/m).to_s.to_f
  eye = @model.latlong_to_point([lon,lat,alt])
  eye.z = alt(eye, mode)
  x = Math.sin(heading.degrees); y = Math.cos(heading.degrees); z = Math.sin(-0.5*Math::PI+tilt.degrees) # z: vorzeichen geändert
  vector = Geom::Vector3d.new x,y,z
  target = eye+vector
  trans_roll = Geom::Transformation.new(eye, [x,y,z], roll.degrees)
  #up = (eye+[0,0,1]).transform!(trans_roll) - eye
  up = [0,0,1].transform!(trans_roll)

  camera = Sketchup::Camera.new eye, target, up
  return camera
end #def camera


def self.lookat(look)
  mode = look.scan(/<altitudeMode>\s*(.*?)\s*<\/altitudeMode>/).to_s
  lon = look.scan(/<longitude>\s*([e0-9\.-]+)\s*<\/longitude>/m).to_s.to_f
  lat = look.scan(/<latitude>\s*([e0-9\.-]+)\s*<\/latitude>/m).to_s.to_f
  alt = look.scan(/<altitude>\s*([e0-9\.-]+)\s*<\/altitude>/m).to_s.to_f
  heading = look.scan(/<heading>\s*([e0-9\.-]+)\s*<\/heading>/m).to_s.to_f
  tilt = look.scan(/<tilt>\s*([e0-9\.-]+)\s*<\/tilt>/m).to_s.to_f
  range = look.scan(/<range>\s*([e0-9\.-]+)\s*<\/range>/m).to_s.to_f
  target = @model.latlong_to_point([lon,lat,alt])
  target.z = alt(target, mode)
  x = Math.sin(heading.degrees); y = Math.cos(heading.degrees); z = Math.sin(-0.5*Math::PI+tilt.degrees) # !!! überprüfen
  vector = Geom::Vector3d.new x,y,z
  vector.length = range.m.to_inch
  eye = target-vector
  up = [0,0,1]

  camera = Sketchup::Camera.new eye, target, up
  return camera
end #def lookat



def self.style(s)
  return {} if s.nil? || s.empty?
  id = s.scan(/<Style\s*?id="(.*?)">/m).to_s
  styl = {}
  linecolor = s.scan(/<LineStyle.*?<color>\s*(.*?)\s*<\/color>.*?<\/LineStyle/m).to_s
  linecolor = self.color_to_su(linecolor)
  styl["linecolor"] = linecolor if linecolor!=nil
  facecolor = s.scan(/<PolyStyle.*?<color>\s*(.*?)\s*<\/color>.*?<\/PolyStyle/m).to_s
  facecolor = self.color_to_su(facecolor)
  styl["facecolor"] = facecolor if facecolor!=nil
  facefill = s.scan(/<PolyStyle.*?<fill>\s*(.*?)\s*<\/fill>.*?<\/PolyStyle/m).to_s
  styl["facefill"] = facefill.to_i if !facefill.empty?
  @styles[id] = styl if id
  return styl
end #def styles



def self.time(data, ent)
  time_begin = data.scan(/<begin>(.*?)<\/begin>/m).to_s
  time_end = data.scan(/<end>(.*?)<\/end>/m).to_s
  time_when = data.scan(/<when>(.*?)<\/when>/m).to_s
  if !time_begin.empty? || !time_end.empty?
    ent.set_attribute("Time","begin",time_begin) if !time_begin.empty?
    ent.set_attribute("Time","end",time_end) if !time_end.empty?
  elsif !time_when.empty?
    ent.set_attribute("Time","when",time_when)
  end
end #def time



##################################################
##########        helper methods        ##########
##################################################



##
# Wrapper loop that updates progressbar
# First set the total amounts of steps to process using:
#   @optdlg.execute_script("startProgress("+n.to_s+")")
# Alternatively, the progressbar can be reset and limited to
# only the steps of this specific loop:
#   startnew=true
#   (for example if the total amount of loops is unknown)
# If Fixnum is given, then that amount of steps is skipped.
#
def self.eachp(a,processname=nil,startnew=false,&block)
  return @dlg.execute_script("ProgressAll-="+a.to_s) if a.class==Fixnum # skip a steps
  @dlg.execute_script("startProgress("+a.length.to_s+")") if startnew==true
  a.each_with_index{|x,i|
    case block.arity
    when 1 then yield x if block_given?
    when 2 then yield x,i if block_given?
    end
    if i=0 && processname!=nil
      @dlg.execute_script("progress('"+processname+"')")
    else
      @dlg.execute_script("progress()")
    end
  }
end




##
# Georeference model
#
def self.set_geo(lon,lat,alt)
  @model.set_attribute("GeoReference","Longitude",lon)
  @model.set_attribute("GeoReference","Latitude",lat)
  @model.set_attribute("GeoReference","Altitude",alt)
  @model.set_attribute("GeoReference","LocationSource","Manual")
  @model.set_attribute("GeoReference","ZValueCentered",-alt.m)
  @model.set_attribute("GeoReference","ModelTranslationZ",-alt.m)
  @model.set_attribute("GeoReference","ModelTranslationY",0) # TODO: SketchUp 7 requires a value here (?), but what does the value mean?
  @model.set_attribute("GeoReference","ModelTranslationX",0) # TODO
  @model.set_attribute("GeoReference", "UsesGeoReferencing", true)
  shadowinfo = @model.shadow_info
  shadowinfo["City"] = "Unknown"
  shadowinfo["Country"] = "Unknown"
  shadowinfo["Longitude"] = lon
  shadowinfo["Latitude"] = lat
end #def set_geo



##
# Calculate altitude of a point, considering altitudeMode.
# (everything in model coordinates)
#
def self.alt(p, mode)
  alt = p.z.to_f
  ter = self.terrain_alt(p)
  if mode == "clampToGround" || mode.length==0
    return ter
  elsif mode == "relativeToGround"
    return alt+ter
  elsif mode == "absolute"
    return alt-@ground.m
  end
  # TODO: relativeToSeaFloor, clampToSeaFloor
end #def alt



##
# Get altitude from Google Earth Terrain
# (in model coordinates)
#
def self.terrain_alt(point)
  p = Geom::Point3d.new(point)
  p.z = -100000
  # raytest:
  # do NOT use [ray]=[point,vector] but point,vector,wysiwyg_flag
  # because SU<8M1 interprets two parameters also as a ray
  raytest = @model.raytest(p,[0,0,1],false)
  until raytest==nil || raytest[1][0].is_a?(Sketchup::Group) && raytest[1][0].name=="Google Earth Terrain"
      p.z = raytest[0].z
      raytest = @model.raytest(p,[0,0,1],false)
  end
  return (!raytest.nil?)? raytest[0].z : 0
end #def terrain_alt



##
# Converts Google Earth's hexadecimal aabbggrr colors to SketchUp's rgba
#
def self.color_to_su(ge)
  return nil if ge.nil? || ge.empty?
  rgba = ge.scan(/../).collect{|x| x.hex}.reverse!
  rgba.push(255) until rgba.length==4
  a = (100*rgba.pop/255.to_f).round/100.to_f
  c = Sketchup::Color.new(rgba)
  m = @model.materials.add ge
  m.color = c
  m.alpha = a
  if @model.materials.respond_to?("remove")
    @model.materials.each{|x| # find identical existing material
      if (m.alpha==x.alpha&&m.color.to_i==x.color.to_i&&m.name!=x.name)
        @model.materials.remove m if Sketchup.version_number >= 8000999 # SU8.0M1
        m = x
      end
    }
  end
  return m
end #def color_to_su



##
#  Calculates the average plane between more or less precise positioned points (or random points)
#
def self.points_to_plane(array)
  a = Array.new(array.compact)
  while a.length > 3
    b = a.shift
    c = a.shift
    d = []
    (0..2).each{|i| d[i] = (b[i]+c[i])/2.to_f }
    a.push(d)
  end
  vector1 = a[0].vector_to(a[1])
  vector2 = a[1].vector_to(a[2])
  # vector2 = vector2.to_a.reverse if vector1.parallel? vector2
  plane = [a[0], vector1*vector2]
  return plane
end



##
# Check if an array of points forms a plane or not.
#
def self.plane?(array)
  plane = [Geom::Point3d.new(array[0]),array[1].vector_to(array[2])]
  (3..array.length-1).each{|i|
    return false if (array[i]-a[0]).on_plane?(plane)==false
  }
  return true
end



##
# Splits an array a at each duplicate's position
# and returns an array b containing all unique parts
#
def self.split_uniq(a)
  b=[]
  i=0;j=1
  until i==a.length
    until j==a.length
      if a[i]==a[j]
        c=a.slice!(0..(j-1))
        b.push(c)
        j=1
      else
        j+=1
      end
    end
    i+=1
    j=i+1
  end
  b.push(a) if a.length!=0
  return b
end #def split_uniq



end #module KmlImport



class KmlImporter < Sketchup::Importer
     def description
       return "Keyhole Markup Language (*.kml)"
     end

     def file_extension
       return "kml"
     end

     def id
       return "com.sketchup.importers.kml_tools"
     end

     def supports_options?
       return false
     end

     def load_file(file_path, status)
       KmlImport.run(file_path)
       return 0 # 0 is the code for a successful import
     end
   end
Sketchup.register_importer(KmlImporter.new)



if( not file_loaded?(File.basename(__FILE__)) )
  cmd_import_kml = UI::Command.new("Import KML File"){KmlImport.run}
  $kml_tools_menu = UI.menu("Plugins").add_submenu("KML Tools") if !$kml_tools_menu
  $kml_tools_menu.add_item(cmd_import_kml)
  file_loaded(File.basename(__FILE__))
end
